from typing import Callable

from .request import Request


class Websocket(Request):
    def __init__(self, scope: dict, receive: Callable, send: Callable):
        assert scope['type'] == 'websocket'
        self.subprotocols = scope.get('subprotocols')
        super().__init__(scope, receive, send)



from http import cookies as http_cookies
from typing import Callable, Optional, Any
from urllib.parse import parse_qsl

import ujson
from starlette.datastructures import UploadFile
from starlette.requests import Request as StarletteRequest

from .multipart import parse_content_header, parse_multipart_form, parse_url_encoded_form_data


def cookie_parser(cookie_string: str) -> dict[str, str]:
    cookie_dict: dict[str, str] = {}
    for chunk in cookie_string.split(";"):
        if "=" in chunk:
            key, val = chunk.split("=", 1)
        else:
            key, val = "", chunk
        key, val = key.strip(), val.strip()
        if key or val:
            cookie_dict[key] = http_cookies._unquote(val)
    return cookie_dict


class Request:
    def __init__(self, scope: dict, receive: Callable, send: Callable) -> None:
        self.scope = scope
        self.receive = receive
        self.send = receive
        self._query_params = None
        self._headers: Optional[dict] = None
        self._path = None
        self._method = None
        self._client = None
        self._host = None
        self._body = None
        self._json = None
        self._files = None
        self._user = None
        self._form = None
        self._session = None
        self._cookies = None
        if scope["type"] == "http":
            self.starlette_request = StarletteRequest(scope, receive, send)

    @property
    def query_params(self) -> dict:
        if self._query_params is None:
            self._query_params = {k.decode(): v.decode() for k, v in
                                  dict(parse_qsl(self.scope["query_string"])).items()}
        return self._query_params

    @property
    def path_params(self) -> dict[str, Any]:
        return self.scope.get("path_params", {})

    @property
    def path(self) -> str:
        if self._path is None:
            self._path = self.scope.get('raw_path').decode()
        return self._path

    @property
    def method(self) -> str:
        if self._method is None:
            self._method = self.scope.get('method')
        return self._method

    @property
    def user(self) -> Any:
        return self._user

    @user.setter
    def user(self, u: Any):
        self._user = u

    @property
    def host(self) -> str:
        if self._host is None:
            host: tuple[str, int] = self.scope.get('server')
            self._host = f"{self.scope.get('scheme')}://{host[0]}:{host[1]}"
        return self._host

    @property
    def headers(self) -> dict:
        if self._headers is None:
            self._headers = {}
            scope_headers: list[tuple[bytes, bytes]] = list(self.scope["headers"])
            for key, value in scope_headers:
                self._headers[key.decode()] = value.decode()
        return self._headers

    @property
    def client(self) -> list:
        if self._client is None:
            self._client = self.scope.get("client", ())
        return self._client

    @property
    def session(self) -> dict:
        if self._session is None:
            self._session = {}
        return self._session

    @session.setter
    def session(self, session: dict):
        self._session = session

    def set_body(self, body: str):
        self._body = body

    async def body(self) -> str:
        if self._body is None:
            body = b""
            while True:
                message = await self.receive()
                body += message.get("body", b"")
                if not message.get("more_body", False):
                    break
            self._body = body
        return self._body.decode()

    async def files(self) -> list[UploadFile]:
        form_data = await self.starlette_request.form()
        return form_data.getlist('files')

    async def json(self) -> dict:
        if self._json is None:
            body = await self.body()
            try:
                self._json = ujson.loads('{}' if body == '' else body)
            except ujson.JSONDecodeError:
                self._json = {}
        return self._json

    @property
    def content_type(self) -> tuple[str, dict[str, str]]:
        _content_type = self.headers.get("content-type", "")
        return parse_content_header(
            _content_type
        )

    async def form(self) -> dict:
        if self._form is None:
            content_type, options = self.content_type
            if content_type == "multipart/form-data":
                self._form = parse_multipart_form(
                    body=(await self.body()).encode(),
                    boundary=options.get("boundary", "").encode(),
                    multipart_form_part_limit=1000,
                )
            elif content_type == "application/x-www-form-urlencoded":
                self._form = parse_url_encoded_form_data(
                    (await self.body()).encode(),
                )
            else:
                self._form = None
        return self._form

    @property
    def cookies(self) -> dict[str, str]:
        if self._cookies is None:
            cookies: dict[str, str] = {}
            cookie_header = self.headers.get("cookie")
            if cookie_header:
                cookies = cookie_parser(cookie_header)
            self._cookies = cookies
        return self._cookies



import dataclasses
import traceback
import typing
from typing import Type, Union

from pydantic import BaseModel

from nestipy.common.exception import HttpException
from nestipy.common.exception.message import HttpStatusMessages
from nestipy.common.exception.status import HttpStatus
from nestipy.common.http_ import Request, Response
from nestipy.common.logger import logger
from nestipy.common.utils import snakecase_to_camelcase
from nestipy.core.exception.processor import ExceptionFilterHandler
from nestipy.core.guards import GuardProcessor
from nestipy.core.interceptor import RequestInterceptor
from nestipy.core.middleware import MiddlewareExecutor
from nestipy.core.template import TemplateRendererProcessor
from nestipy.ioc import NestipyContainer
from nestipy.ioc import RequestContextContainer
from nestipy.types_ import NextFn, CallableHandler
from .route_explorer import RouteExplorer
from ..adapter.http_adapter import HttpAdapter
from ..context.execution_context import ExecutionContext
from ...openapi.openapi_docs.v3 import Operation, PathItem


class RouterProxy:
    def __init__(self, router: HttpAdapter):
        self.router = router
        self._template_processor = TemplateRendererProcessor(router)

    def apply_routes(self, modules: list[Union[Type, object]]):
        json_paths = {}
        for module_ref in modules:
            routes = RouteExplorer.explore(module_ref)
            for route in routes:
                path = route['path']
                methods = route['request_method']
                method_name = route['method_name']
                controller = route['controller']
                handler = self.create_request_handler(module_ref, controller, method_name)
                for method in methods:
                    getattr(self.router, method.lower())(path, handler, route)

                    # OPEN API REGISTER
                    if path in json_paths.keys():
                        route_path = json_paths[path]
                    else:
                        route_path = {}
                    if "responses" not in route['openapi'].keys():
                        continue

                    if 'no_swagger' not in route['openapi'].keys():
                        route_path[method.lower()] = Operation(
                            **route['openapi'],
                            summary=snakecase_to_camelcase(method_name)
                        )
                        json_paths[path] = route_path
        paths = {}
        for path, op in json_paths.items():
            paths[path] = PathItem(**op)
        return paths

    def create_request_handler(
            self,
            module_ref: Type,
            controller: Union[object, Type],
            method_name: str
    ) -> CallableHandler:

        async def request_handler(req: "Request", res: "Response", next_fn: NextFn):

            context_container = RequestContextContainer.get_instance()
            # setup container for query params, route params, request, response, session, etc..
            # await self.setup_context_data(context_container, req, res)

            # TODO: req.files

            container = NestipyContainer.get_instance()
            controller_method_handler = getattr(controller, method_name)
            execution_context = ExecutionContext(
                self.router,
                module_ref,
                controller,
                controller_method_handler,
                req,
                res
            )
            context_container.set_execution_context(execution_context)
            handler_response: Response
            try:
                # TODO : Refactor
                guard_processor: GuardProcessor = await NestipyContainer.get_instance().get(GuardProcessor)
                can_activate = await guard_processor.process(execution_context)
                if not can_activate[0]:
                    # Raise error
                    raise HttpException(
                        HttpStatus.UNAUTHORIZED,
                        HttpStatusMessages.UNAUTHORIZED,
                        details=f"Not authorized from guard {can_activate[1]}"
                    )

                # create next_function that call catch
                async def next_fn_middleware(ex: typing.Any = None):
                    if ex is not None:
                        return await self._ensure_response(res, await next_fn(ex))
                    return await container.get(controller, method_name)

                async def next_fn_interceptor(ex: typing.Any = None):
                    if ex is not None:
                        return await self._ensure_response(res, await next_fn(ex))
                    return await MiddlewareExecutor(req, res, next_fn_middleware).execute()

                #  execute Interceptor by using middleware execution as next_handler
                interceptor: RequestInterceptor = await container.get(RequestInterceptor)
                result = await interceptor.intercept(execution_context, next_fn_interceptor)
                if result is None:
                    raise HttpException(
                        HttpStatus.BAD_REQUEST,
                        "Handler not called because of interceptor: Invalid Request"
                    )

                # process template rendering
                if self._template_processor.can_process(controller_method_handler, result):
                    result = await res.html(self._template_processor.render())
                # transform result to response
                handler_response = await self._ensure_response(res, result)

            except Exception as e:
                tb = traceback.format_exc()
                logger.error(e)
                logger.error(tb)
                if not isinstance(e, HttpException):
                    e = HttpException(HttpStatus.INTERNAL_SERVER_ERROR, str(e), str(tb))

                # Call exception catch
                exception_handler = await container.get(ExceptionFilterHandler)
                result = await exception_handler.catch(e, execution_context)
                if result:
                    handler_response = await self._ensure_response(res, result)
                else:
                    handler_response = await self._ensure_response(res, await next_fn(e))
            finally:
                #  reset request context container
                context_container.destroy()
            return handler_response

        return request_handler

    @classmethod
    async def _ensure_response(cls, res: "Response", result: Union["Response", str, dict, list]) -> "Response":

        if isinstance(result, (str, int, float)):
            return await res.send(content=str(result), status_code=200)
        elif isinstance(result, (list, dict)):
            return await res.json(content=result, status_code=200)
        elif dataclasses.is_dataclass(result):
            return await res.json(
                content=dataclasses.asdict(typing.cast(dataclasses.dataclass, result)),
                status_code=200
            )
        elif isinstance(result, BaseModel):
            return await res.json(content=result.dict(), status_code=200)
        elif isinstance(result, Response):
            return result
        else:
            return await res.json(content={'error': 'Unknown response format'}, status_code=403)



from dataclasses import asdict
from typing import Literal, Annotated

from nestipy.common.decorator import Module, Injectable
from nestipy.dynamic_module import ConfigurableModuleBuilder
from nestipy.ioc import ModuleProviderDict, Inject
from nestipy.metadata import Reflect

Config = dict[Literal['folder'], str]
ConfigurableModuleClass, CONFIG_MODULE_OPTION_TOKEN = ConfigurableModuleBuilder[Config]().set_method('for_root').build()


@Injectable()
class ConfigService:
    token: Annotated[str, Inject(CONFIG_MODULE_OPTION_TOKEN)]


@Module(
    providers=[
        ConfigService,
        ModuleProviderDict(value='hello', token='MyProvider')
    ],
    exports=[
        ConfigService
    ],
    is_global=True
)
class ConfigModule(ConfigurableModuleClass):
    ...


if __name__ == '__main__':
    d = ConfigModule.for_root({'folder': './config'})
    meta = Reflect.get(d)
    print(meta, {k: v for k, v in asdict(d).items() if k != 'module'})

__all__ = [
    "ConfigService",
    "ConfigModule"
]



"""
This module contains classes that are specific to the OpenAPI Documentation
Specification V3, to create examples of values from schemas. These are used to produce
basic examples of request and response bodies from schemas, when specific examples are
not included in documentation files.
"""
import random
from abc import ABC, abstractmethod
from typing import Any, Callable, ClassVar, Dict, Iterable, List, Type
from uuid import uuid4


class SchemaExampleHandler(ABC):
    """
    Base type for classes that can generate basic example values
    from schema types.
    """

    type_name: ClassVar[str]

    @abstractmethod
    def get_example(self, schema) -> Any:
        """
        Returns an example value for a property with the given name and schema.
        """


# TODO: allOf, anyOf, oneOf, not (?)
# https://swagger.io/docs/specification/data-models/oneof-anyof-allof-not/


class ScalarExampleHandler(SchemaExampleHandler):
    type_name = ""
    default: Any
    formats: Dict[str, Callable[[], Any]]

    def get_example(self, schema) -> str:
        format = schema.get("format")

        if format and format in self.formats:
            return self.formats[format]()

        return self.default


class StringExampleHandler(ScalarExampleHandler):
    type_name = "string"
    default = "string"
    formats = {
        "email": lambda: "derp@meme.org",
        "uuid": lambda: str(uuid4()),
        "date": lambda: "2022-04-13",
        "date-time": lambda: "2022-04-13T15:42:05.901Z",
        "password": lambda: "*" * 12,
        "byte": lambda: "TG9yZW0gaXBzdW0gZG9sb3Igc2l0IGFtZXQ=",
        "binary": lambda: "TG9yZW0gaXBzdW0gZG9sb3Igc2l0IGFtZXQ=",
    }

    def get_example(self, schema) -> str:
        enum = schema.get("enum")
        if isinstance(enum, list):
            return enum[0]
        return super().get_example(schema)


class IntegerExampleHandler(ScalarExampleHandler):
    type_name = "integer"
    default = 0
    formats = {
        "int32": lambda: random.randint(0, 300),
        "int64": lambda: random.randint(0, 300),
    }


class BooleanExampleHandler(ScalarExampleHandler):
    type_name = "boolean"
    default = True
    formats = {}


class NumberExampleHandler(ScalarExampleHandler):
    type_name = "number"
    default = 10.12
    formats = {
        "float": lambda: 10.12,
        "double": lambda: 10.12,
    }


class ObjectExampleHandler(SchemaExampleHandler):
    type_name = "object"

    def get_example(self, schema) -> Any:
        """
        Returns an example value for a property with the given name and schema.
        """
        properties = schema.get("properties") or {}

        example = {}

        for key in properties:
            example[key] = get_example_from_schema(properties[key])

        return example


class ArrayExampleHandler(SchemaExampleHandler):
    type_name = "array"

    def get_example(self, schema) -> Any:
        """
        Returns an example value for a property with the given name and schema.

        Example:
          type: array
          items:
            $ref: '#/components/schemas/ReleaseNodeDownload'
          nullable: true
        """
        items = schema.get("items", [])

        if not isinstance(items, list):
            items = [items]

        return [get_example_from_schema(item) for item in items]


def get_subclasses(cls) -> Iterable[Type]:
    for subclass in cls.__subclasses__():
        yield from get_subclasses(subclass)
        yield subclass


def get_example_from_schema(schema) -> Any:
    if schema is None:
        return None

    if "example" in schema:
        return schema["example"]

    # does it have a type?
    handlers_types: List[Type[SchemaExampleHandler]] = list(
        get_subclasses(SchemaExampleHandler)
    )

    schema_type = schema.get("type")

    if schema_type:
        handler_type = next(
            (_type for _type in handlers_types if _type.type_name == schema_type), None
        )

        if handler_type is None:  # pragma: nocover
            # fallback to returning the raw schema;
            return schema

        handler = handler_type()
        return handler.get_example(schema)
    # TODO: handle special cases (allOf, anyOf, etc.)
    return None



from typing import Callable

import uvicorn

from request import Request
from response import Response


async def app(scope: dict, receive: Callable, send: Callable):
    if scope.get('type') == 'http':
        request = Request(scope, receive, send)
        json = await request.json()
        print(request.query_params, json)
        res = Response()
        await res.status(200).json({'test': 'test'})


if __name__ == '__main__':
    uvicorn.run('test:app', reload=True)



from typing import Any, Union, Type, Callable


class Reflect:
    metadata: str = '__reflect__metadata__'

    @classmethod
    def set_metadata(cls, obj: Union[Type, object, Callable], key: str, value: Any):
        meta = getattr(obj, cls.metadata, {})
        meta[key] = value
        setattr(obj, cls.metadata, meta)

    @classmethod
    def get_metadata(cls, obj: Union[Type, object, Callable], key: str, default: Any = None):
        meta = getattr(obj, cls.metadata, {})
        if key not in meta.keys():
            return default
        return meta[key]

    @classmethod
    def get(cls, obj: Union[Type, object, Callable], ):
        return getattr(obj, cls.metadata, {})



from __future__ import annotations

import re
from collections import defaultdict
from email.utils import decode_rfc2231
from functools import lru_cache
from typing import Any
from urllib.parse import unquote, parse_qsl as _parse_qsl

from .upload_file import UploadFile

__all__ = ("parse_body", "parse_content_header", "parse_multipart_form", "parse_url_encoded_form_data")

_token = r"([\w!#$%&'*+\-.^_`|~]+)"  # noqa: S105
_quoted = r'"([^"]*)"'
_param = re.compile(rf";\s*{_token}=(?:{_token}|{_quoted})", re.ASCII)
_firefox_quote_escape = re.compile(r'\\"(?!; |\s*$)')


def parse_qsl(qs: bytes, separator: str) -> list[tuple[str, str]]:
    return _parse_qsl(qs.decode("latin-1"), keep_blank_values=True, separator=separator)


@lru_cache(1024)
def parse_url_encoded_form_data(encoded_data: bytes) -> dict[str, str | list[str]]:
    """Parse an url encoded form data dict.

    Args:
        encoded_data: The encoded byte string.

    Returns:
        A parsed dict.
    """
    decoded_dict: defaultdict[str, list[str]] = defaultdict(list)
    for k, v in parse_qsl(encoded_data, separator="&"):
        decoded_dict[k].append(v)
    return {k: v if len(v) > 1 else v[0] for k, v in decoded_dict.items()}


def parse_content_header(value: str) -> tuple[str, dict[str, str]]:
    value = _firefox_quote_escape.sub("%22", value)
    pos = value.find(";")
    if pos == -1:
        options: dict[str, str] = {}
    else:
        options = {
            m.group(1).lower(): m.group(2) or m.group(3).replace("%22", '"') for m in _param.finditer(value[pos:])
        }
        value = value[:pos]
    return value.strip().lower(), options


def parse_body(body: bytes, boundary: bytes, multipart_form_part_limit: int) -> list[bytes]:
    if not (body and boundary):
        return []

    form_parts = body.split(boundary, multipart_form_part_limit + 3)[1:-1]

    if len(form_parts) > multipart_form_part_limit:
        raise Exception(
            f"number of multipart components exceeds the allowed limit of {multipart_form_part_limit}, "
            f"this potentially indicates a DoS attack"
        )

    return form_parts


def parse_multipart_form(
        body: bytes,
        boundary: bytes,
        multipart_form_part_limit: int = 1000,
        type_decoders: Any | None = None,
) -> dict[str, Any]:
    fields: defaultdict[str, list[Any]] = defaultdict(list)

    for form_part in parse_body(body=body, boundary=boundary, multipart_form_part_limit=multipart_form_part_limit):
        file_name = None
        content_type = "text/plain"
        content_charset = "utf-8"
        field_name = None
        line_index = 2
        line_end_index = 0
        headers: list[tuple[str, str]] = []

        while line_end_index != -1:
            line_end_index = form_part.find(b"\r\n", line_index)
            form_line = form_part[line_index:line_end_index].decode("utf-8")

            if not form_line:
                break

            line_index = line_end_index + 2
            colon_index = form_line.index(":")
            current_idx = colon_index + 2
            form_header_field = form_line[:colon_index].lower()
            form_header_value, form_parameters = parse_content_header(form_line[current_idx:])

            if form_header_field == "content-disposition":
                field_name = form_parameters.get("name")
                file_name = form_parameters.get("filename")

                if file_name is None and (filename_with_asterisk := form_parameters.get("filename*")):
                    encoding, _, value = decode_rfc2231(filename_with_asterisk)
                    file_name = unquote(value, encoding=encoding or content_charset)

            elif form_header_field == "content-type":
                content_type = form_header_value
                content_charset = form_parameters.get("charset", "utf-8")
            headers.append((form_header_field, form_header_value))

        if field_name:
            post_data = form_part[line_index:-4].lstrip(b"\r\n")
            if file_name:
                form_file = UploadFile(
                    content_type=content_type, filename=file_name, file_data=post_data, headers=dict(headers)
                )
                fields[field_name].append(form_file)
            elif post_data:
                fields[field_name].append(post_data.decode(content_charset))
            else:
                fields[field_name].append(None)

    return {k: v if len(v) > 1 else v[0] for k, v in fields.items()}



from __future__ import annotations

from tempfile import SpooledTemporaryFile
from typing import Any

from pydantic_core import core_schema

__all__ = ("UploadFile",)


class UploadFile:

    def __init__(
            self,
            content_type: str,
            filename: str,
            file_data: bytes | None = None,
            headers: dict[str, str] | None = None,
            max_spool_size: int = 1024 * 1024,
    ) -> None:
        self.filename = filename
        self.content_type = content_type
        self.file = SpooledTemporaryFile(max_size=max_spool_size)
        self.headers = headers or {}

        if file_data:
            self.file.write(file_data)
            self.file.seek(0)

    @property
    def rolled_to_disk(self) -> bool:
        return False

    async def write(self, data: bytes) -> int:
        return self.file.write(data)

    async def read(self, size: int = -1) -> bytes:
        return self.file.read(size)

    async def seek(self, offset: int) -> int:
        return self.file.seek(offset)

    async def close(self) -> None:
        return self.file.close()

    def __repr__(self) -> str:
        return f"{self.filename} - {self.content_type}"

    @classmethod
    def __get_pydantic_core_schema__(
            cls,
            _source_type: Any,
            _handler: Any,
    ) -> core_schema.CoreSchema:
        return core_schema.bytes_schema()



import typing

from nestipy.metadata import Reflect


class OpenApiExplorer:

    @classmethod
    def explore(cls, obj):
        deps: dict = {}
        meta = Reflect.get(obj)
        for key, value in meta.items():
            if typing.cast(str, key).startswith('__openapi__'):
                openapi_key = typing.cast(str, key).replace('__openapi__', '')
                deps[openapi_key] = value

        return deps



class ModuleMetadata:
    Imports: str = 'imports'
    Exports: str = 'exports'
    Providers: str = 'providers'
    Controllers: str = 'controllers'
    Global: str = '_global_'
    Root: str = '_root_'
    Module: str = '_is_module_'



import os
from typing import Union, TYPE_CHECKING, AsyncIterator, Callable

import aiofiles
import ujson as json

from nestipy.common.exception.http import HttpException
from nestipy.common.exception.message import HttpStatusMessages
from nestipy.common.exception.status import HttpStatus

if TYPE_CHECKING:
    from nestipy.common.template import TemplateEngine


class Response:
    _status_code: int = 200
    _headers: set[tuple[str, str]] = set()
    _cookies: set[tuple[str, str]] = set()
    _content: Union[bytes, None]

    def __init__(self, template_engine: "TemplateEngine" = None) -> None:
        self._headers = set()
        self._status_code = 200
        self._content = None
        self._stream_content = None
        self.template_engine = template_engine

    async def _start(self) -> None:
        # await self._send({
        #     "type": "http.response.start",
        #     "status": self._status_code,
        #     "headers": list(self._headers),
        # })
        pass

    async def _write(self, content: bytes) -> None:
        # await self._start()
        # await self._send({
        #     "type": "http.response.body",
        #     "body": content,
        #     "more_body": False,
        # })
        self._content = content

    def status(self, status_code: int) -> "Response":
        self._status_code = status_code
        return self

    def header(self, name: str, value: str) -> "Response":
        self._headers = set([(k, v) for k, v in self._headers if k.lower() != name.lower])
        self._headers.add((name, value))
        return self

    def cookie(self, name: str, value: str) -> "Response":
        self._cookies = set([(k, v) for k, v in self._cookies if k.lower() != name.lower])
        self._cookies.add((name, value))
        return self

    async def send(self, content: str, status_code: int = 200) -> "Response":
        await self.status(status_code)._write(content.encode())
        return self

    async def html(self, content: str, status_code: int = None):
        self.header("Content-Type", "text/html")
        self.header('max-age', '0')
        if status_code is not None:
            self.status(status_code)
        await self._write(content.encode())
        return self

    async def redirect(self, url: str, status_code: int = 302) -> "Response":
        self.status(status_code)
        self.header("Location", url)
        await self._write(b"")
        return self

    async def json(self, content: dict, status_code: int = None) -> "Response":
        if status_code is not None:
            self.status(status_code)
        self.header("Content-Type", "application/json")
        await self._write(json.dumps(content).encode("utf-8"))
        return self

    async def download(self, file_path: str, file_name: str, attachment: bool = True) -> "Response":
        if not os.path.isfile(file_path):
            self.status(404)
            await self._write(b"File not found")
            return self

        self.header("Content-Disposition", f"{'attachment; ' if attachment else ''}filename={file_name}")
        async with aiofiles.open(file_path, "rb") as file:
            content = await file.read()
            self.header("Content-Length", str(len(content)))
            await self._write(content)
        return self

    async def stream(self, callback: Callable[[], AsyncIterator[Union[bytes, str]]]):

        self._stream_content = callback
        return self

    async def stream_file(self, file_path: str, chunk_size: 4096) -> "Response":
        if not os.path.isfile(file_path):
            self.status(404)
            await self._write(b"File not found")
            return self

        async def get_stream() -> AsyncIterator[bytes]:
            async with aiofiles.open(file_path, "rb") as file:
                chunk: bytes = await file.read(chunk_size)
                # no benefit because we don't send response directly
                while chunk:
                    yield chunk
                    chunk: bytes = await file.read(chunk_size)

        self._stream_content = get_stream

        return self

    async def render(self, template: str, context=None):
        if context is None:
            context = {}
        if self.template_engine is not None:
            content = self.template_engine.render(template, context)
            await self.html(content)
        else:
            raise HttpException(
                HttpStatus.BAD_GATEWAY,
                HttpStatusMessages.BAD_GATEWAY,
                'Template engine not configured'
            )
        return self

    def content_type(self) -> str:
        for key, value in self._headers:
            if str(key).lower() == 'content-type':
                return value
        return 'text/plain'

    def status_code(self) -> int:
        return self._status_code

    def content(self) -> Union[bytes, None]:
        return self._content

    def headers(self) -> set[tuple[str, str]]:
        return self._headers

    def cookies(self) -> set[tuple[str, str]]:
        return self._cookies

    def is_stream(self) -> bool:
        return self._stream_content is not None

    async def get_stream(self) -> AsyncIterator[bytes]:
        if self._stream_content is not None:
            yield self._stream_content()



from typing import Type, Union, Callable, Any

from nestipy.metadata import Reflect

from nestipy.graphql.meta import NestipyGraphqlKey


class GraphQlDecorator:
    def __init__(self, graphql_type: str, value: Any = True, is_method: bool = True):
        self.graphql_type = graphql_type
        self.graphql_value = value
        self.is_method = is_method

    def __call__(self, return_type: Type = None, *args, **kwargs):
        def wrapper(class_ref: Union[Type, Callable]):
            from nestipy.common.decorator import Injectable
            if not self.is_method:
                class_ref = Injectable()(class_ref)
            Reflect.set_metadata(class_ref, self.graphql_type, self.graphql_value)
            Reflect.set_metadata(class_ref, NestipyGraphqlKey.return_type, return_type)
            return class_ref

        return wrapper


def Resolver(return_type: Type = None, *args, **kwargs):
    return GraphQlDecorator(NestipyGraphqlKey.resolver, is_method=False)(return_type, *args, **kwargs)


def Query(return_type: Type = None, *args, **kwargs):
    return GraphQlDecorator(NestipyGraphqlKey.handler, NestipyGraphqlKey.query)(return_type, *args, **kwargs)


def Mutation(return_type: Type = None, *args, **kwargs):
    return GraphQlDecorator(NestipyGraphqlKey.handler, NestipyGraphqlKey.mutation)(return_type, *args, **kwargs)


def Subscription(return_type: Type = None, *args, **kwargs):
    return GraphQlDecorator(NestipyGraphqlKey.handler, NestipyGraphqlKey.subscription)(return_type, *args, **kwargs)



"""
This module contains classes to generate representations of content types by mime type.
"""
import os
from abc import ABC, abstractmethod
from datetime import datetime
from json import JSONEncoder
from urllib.parse import urlencode

from essentials.json import FriendlyEncoder, dumps


class OADJSONEncoder(JSONEncoder):
    def default(self, obj):
        try:
            return JSONEncoder.default(self, obj)
        except TypeError:
            if isinstance(obj, datetime):
                datetime_format = os.environ.get("OPENAPI_DATETIME_FORMAT")
                if datetime_format:
                    return obj.strftime(datetime_format)
                else:
                    return obj.isoformat()
            return FriendlyEncoder.default(self, obj)  # type: ignore


class ContentWriter(ABC):
    """
    Base type for classes that can create representations of request/response contents.
    """

    @abstractmethod
    def handle_content_type(self, content_type: str) -> bool:
        """
        Returns a value indicating whether this example handler can handle an example
        of the given type.
        """

    @abstractmethod
    def write(self, value) -> str:
        """
        Writes markdown to represent a value in a certain type of content.
        """


class JSONContentWriter(ContentWriter):
    def handle_content_type(self, content_type: str) -> bool:
        return "json" in content_type.lower()

    def write(self, value) -> str:
        return dumps(value, indent=4, cls=OADJSONEncoder)


class FormContentWriter(ContentWriter):
    def handle_content_type(self, content_type: str) -> bool:
        # multipart/form-data. Otherwise, use application/x-www-form-urlencoded.
        return "x-www-form-urlencoded" == content_type.lower()

    def write(self, value) -> str:
        return urlencode(value)



from blacksheep import Application, Response as BlackSheepResponse
from blacksheep import get, put, post, patch, head, options, delete, route as all_route, Content, ws as websocket, \
    WebSocket as BSWebSocket, StreamedContent
from nestipy.common.http_ import Response
from nestipy.types_ import CallableHandler, WebsocketHandler, MountHandler

from .http_adapter import HttpAdapter


class BlackSheepAdapter(HttpAdapter):

    def __init__(self):
        self.instance = Application()
        self.instance.on_start(self.on_startup)
        self.instance.on_stop(self.on_shutdown)

    def get_instance(self) -> Application:
        return self.instance

    def engine(self, args, *kwargs) -> None:
        pass

    def enable_cors(self) -> None:
        self.instance.use_cors(
            allow_methods="GET POST PUT DELETE OPTIONS",
            allow_origins="*",
            allow_headers="Content-Type",
            max_age=300,
        )

    def create_wilchard(self, prefix: str = '/', name: str = 'full_path') -> str:
        return f"/{prefix.strip('/')}" + "/{" + f"path:{name}" + "}"

    def use(self, callback: CallableHandler, metadata: dict) -> None:
        # need to transform
        self.instance.middlewares.append(callback)

    def get(self, route: str, callback: CallableHandler, metadata: dict) -> None:
        get(route)(self._create_blacksheep_handler(callback, metadata))

    def post(self, route: str, callback: CallableHandler, metadata: dict) -> None:
        post(route)(self._create_blacksheep_handler(callback, metadata))

    def put(self, route: str, callback: CallableHandler, metadata: dict) -> None:
        put(route)(self._create_blacksheep_handler(callback, metadata))

    def delete(self, route: str, callback: CallableHandler, metadata: dict) -> None:
        delete(route)(self._create_blacksheep_handler(callback, metadata))

    def options(self, route: str, callback: CallableHandler, metadata: dict) -> None:
        options(route)(self._create_blacksheep_handler(callback, metadata))

    def head(self, route: str, callback: CallableHandler, metadata: dict) -> None:
        head(route)(self._create_blacksheep_handler(callback, metadata))

    def patch(self, route: str, callback: CallableHandler, metadata: dict) -> None:
        patch(route)(self._create_blacksheep_handler(callback, metadata))

    def all(self, route: str, callback: CallableHandler, metadata: dict) -> None:
        all_route(route)(self._create_blacksheep_handler(callback, metadata))

    def ws(self, route: str, callback: WebsocketHandler, metadata: dict) -> None:
        websocket(route)(self._create_blacksheep_websocket_handler(callback, metadata))

    def mount(self, route: str, callback: MountHandler) -> None:
        self.instance.mount(route, callback)

    def _create_blacksheep_handler(self, callback: CallableHandler, _metadata: dict):
        # path = metadata['path']
        # params = RouteParamsExtractor.extract_params(path)
        async def blacksheep_handler() -> BlackSheepResponse:
            req, res, next_fn = self.create_handler_parameter()
            result: Response = await callback(req, res, next_fn)
            if result.is_stream():
                return BlackSheepResponse(
                    content=StreamedContent(
                        content_type=result.content_type().encode(),
                        data_provider=result.get_stream
                    ),
                    headers=[(k.encode(), v.encode()) for k, v in result.headers()],
                    status=result.status_code() or 200
                )
            return BlackSheepResponse(
                content=Content(data=result.content(), content_type=result.content_type().encode()),
                headers=[(k.encode(), v.encode()) for k, v in result.headers()],
                status=result.status_code() or 200
            )

        return blacksheep_handler

    def _create_blacksheep_websocket_handler(self, callback: WebsocketHandler, metadata: dict):
        async def blacksheep_websocket_handler(bsw: BSWebSocket):
            wbs = self.create_websocket_parameter()
            return await callback(wbs)

        return blacksheep_websocket_handler



from typing import Union, Callable, Any, TYPE_CHECKING

from nestipy.metadata import Reflect

from nestipy.common.template import TemplateEngine, TemplateKey
from .minimal_jinja import MinimalJinjaTemplateEngine

if TYPE_CHECKING:
    from nestipy.core.adapter.http_adapter import HttpAdapter


class TemplateRendererProcessor:
    def __init__(self, adapter: "HttpAdapter"):
        self.template = None
        self.adapter = adapter
        self.template_engine = None
        self.context = {}

    def can_process(self, method: Callable, context: Any) -> bool:
        self.template_engine: Union[TemplateEngine, None] = self.adapter.get_state(TemplateKey.MetaEngine)
        self.template = Reflect.get_metadata(method, TemplateKey.MetaRender, None)
        self.context = context
        if self.template_engine is None:
            return False
            # raise HttpException(HttpStatus.INTERNAL_SERVER_ERROR, 'Template engine not configured')
        if self.template is None or not isinstance(context, dict):
            return False
        else:
            return True

    def render(self) -> Union[str, None]:
        template_engine: Union[TemplateEngine, None] = self.adapter.get_state(TemplateKey.MetaEngine)
        if template_engine is None:
            return None
        else:
            return template_engine.render(self.template, self.context)


__all__ = [
    "MinimalJinjaTemplateEngine",
    "TemplateRendererProcessor"
]



import os
from abc import abstractmethod

import aiofiles

from nestipy.ioc import RequestContextContainer

from nestipy.common.http_ import Request
from nestipy.graphql.graphql_module import GraphqlOption


class GraphqlAsgi:
    option: GraphqlOption = None

    def set_graphql_option(self, option: GraphqlOption):
        self.option = option

    async def get_graphql_ide(self) -> str:
        playground_path = os.path.join(
            os.path.dirname(__file__),
            'playground',
            f'graphql-playground-{self.option.ide}.html'
        )
        file = await aiofiles.open(playground_path, 'r')
        content = await file.read()
        await file.close()
        return content

    @abstractmethod
    async def handle(self, scope: dict, receive, send):
        self._setup_request(scope, receive, send)

    @classmethod
    def _setup_request(cls, scope: dict, receive, send):
        pass
        # from nestipy.core import ExecutionContext
        # req = Request(scope, receive, send)
        # req_container = RequestContextContainer.get_instance()
        # res = req_container.execution_context.get_response()
        # class_handler = req_container.execution_context.get_class()
        # handler = req_container.execution_context.get_handler()
        # module = req_container.execution_context.get_module()
        # adapter = req_container.execution_context.get_adapter()
        # graphql = req_container.execution_context.switch_to_graphql()
        # req_container.set_execution_context(ExecutionContext(
        #     adapter,
        #     module,
        #     class_handler,
        #     handler,
        #     req,
        #     res,
        #     graphql.get_context(),
        #     graphql.get_args()
        # ))



from typing import Union


class HttpException(Exception):
    def __init__(self, status_code: int, message: str = None, details: Union[dict, str] = None):
        self.status_code = status_code
        self.message = message
        self.details = details
        super().__init__(self.message)

    def __str__(self):
        return f"{self.status_code} - {self.message} {f' - {self.details}' if self.details is not None else ''}"



from abc import ABC, abstractmethod
from typing import Any, Callable

from socketio import AsyncServer

from .socket_request import Websocket


class IoAdapter(ABC):

    def __init__(self, path: str = 'socket.io'):
        self._path = f"/{path.strip('/')}"

    @abstractmethod
    def on(self, event: str, namespace: str = None) -> Callable[[Callable], Any]:
        pass

    @abstractmethod
    def emit(
            self,
            event: Any,
            data: Any = None,
            to: Any = None,
            room: Any = None,
            skip_sid: Any = None,
            namespace: Any = None,
            callback: Any = None,
            ignore_queue: bool = False
    ):
        pass

    @abstractmethod
    def on_connect(self) -> Callable[[Callable], Any]:
        pass

    @abstractmethod
    def on_disconnect(self) -> Callable[[Callable], Any]:
        pass

    @abstractmethod
    def broadcast(self, event: Any, data: Any):
        pass

    @abstractmethod
    async def __call__(self, scope: dict, receive: Callable, send: Callable) -> bool:
        pass


class SocketIoAdapter(IoAdapter):

    def __init__(self, io: AsyncServer, path: str = 'socket.io'):
        super().__init__(path=path)
        self._io = io
        self._connected = []

    def on(self, event: str, namespace: str = None):
        def decorator(handler: Callable):
            async def wrapper(sid: str, data: Any):
                environ = self._io.get_environ(sid, namespace)
                client = Websocket(
                    namespace,
                    sid,
                    data,
                    environ['asgi.scope'],
                    environ['asgi.receive'],
                    environ['asgi.send']
                )
                return await handler(event, client, data)

            self._io.on(event, namespace=namespace)(wrapper)

        return decorator

    async def emit(
            self,
            event: Any,
            data: Any = None,
            to: Any = None,
            room: Any = None,
            skip_sid: Any = None,
            namespace: Any = None,
            callback: Any = None,
            ignore_queue: bool = False):
        return await self._io.emit(event, data, to, room, skip_sid, namespace, callback, ignore_queue)

    def broadcast(self, event: Any, data: Any):
        return self._io.emit(event, data, self._connected)

    def on_connect(self):
        def decorator(handler: Callable):
            async def wrapper(sid: Any, *args, **kwargs):
                self._connected.append(sid)
                return await handler(sid, *args, **kwargs)

            return self._io.on('connect')(wrapper)

        return decorator

    def on_disconnect(self):
        def decorator(handler: Callable):
            async def wrapper(sid: Any, *args, **kwargs):
                self._connected.remove(sid)
                return await handler(sid, *args, **kwargs)

            return self._io.on('disconnect')(wrapper)

        return decorator

    async def __call__(self, scope: dict, receive: Callable, send: Callable):
        if scope['type'] in ['http', 'websocket'] and \
                scope['path'].startswith(self._path):
            await self._io.handle_request(scope, receive, send)
            return True
        return False



from abc import abstractmethod, ABC
from typing import Awaitable, Union, TYPE_CHECKING
if TYPE_CHECKING:
    from nestipy.core.context.execution_context import ExecutionContext


class CanActivate(ABC):

    @abstractmethod
    def can_activate(self, context: "ExecutionContext") -> Union[Awaitable[bool], bool]:
        pass



import asyncio
from typing import Any

from pyee import EventEmitter


class PubSub:
    def __init__(self):
        self.ee = EventEmitter()

    async def asyncIterator(self, event: str):
        loop = asyncio.get_event_loop()
        future = loop.create_future()

        def handler(value):
            loop.call_soon_threadsafe(future.set_result, value)

        self.ee.on(event, handler)
        try:
            while True:
                yield await future
                future = loop.create_future()  # Reset the future for the next value
        finally:
            self.ee.remove_listener(event, handler)

    def publish(self, event: str, data: Any):
        self.ee.emit(event, data)
